#include "PCIDump.h"

PDRIVER_OBJECT gDriverObject = NULL;
PVOID BIOS_VA = NULL;
PVOID SD_Entry = NULL;
PVOID PCI_BIOS = NULL;
ULONG BIOS_ver = -1;
ULONG IrqBitMap = 0;
BOOLEAN mmMapped = FALSE;

NTSTATUS
PciBios32_Init()
{
    ULONG i;
    PBIOS32_SD sd;
    PHYSICAL_ADDRESS a;

    a.QuadPart = 0;
    BIOS_VA = (PVOID)MmMapIoSpace(a, BASE_1MB_SIZE, MmCached);
    if(BIOS_VA) {
        mmMapped = TRUE;
    } else {
        PciBios32_Deinit();
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    
    KdPrint(("New BIOS_VA %x\n", BIOS_VA));
    if(!BIOS_VA) {
        PciBios32_Deinit();
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    for(i=0; i<BASE_BIOS_SIZE; i+=16) {
        sd = Check_BIOS32_SD((ULONG)BIOS_VA + BASE_BIOS_OFFSET + i);
        if(sd) {
            KdPrint(("Found BIOS32 Service Directory descriptor at %x\n", sd));
            break;
        }
    }
    if(!sd) {
        KdPrint(("Can't find BIOS32 Service Directory descriptor\n"));
        PciBios32_Deinit();
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    SD_Entry = (PVOID)((ULONG)BIOS_VA + sd->phAddr);
    KdPrint(("BIOS32 Service Directory entry %x\n", SD_Entry));

    __asm {
        pushad
        xor   ebx,ebx
        mov   ax,cs
        push  eax
        mov   eax,PCI_BIOS32_SD_SIG
        call  SD_Entry
        or    al,al
        jnz   no_pci_bios
        mov   eax,BIOS_VA
        add   eax,ebx
        add   eax,edx
        mov   PCI_BIOS,eax
no_pci_bios:
        popad
    }
    if(PCI_BIOS) {
        KdPrint(("Found PCI BIOS32 entry at %x\n", PCI_BIOS));
    } else {
        KdPrint(("Can't find PCI BIOS32 entry\n"));
        PciBios32_Deinit();
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    BIOS_ver = PciIsBIOSPresent();
    if(BIOS_ver == -1) {
        KdPrint(("PCI BIOS32 presence check failed\n"));
        PciBios32_Deinit();
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    KdPrint(("PCI BIOS32 v%x.%2.2x\n", (BIOS_ver >> 8) & 0xff, BIOS_ver & 0xff));
/*
    IrqRoutingTable = PciGetIrqRoutingTable(&szIrqRoutingTable, &IrqBitMap);
    if(!IrqRoutingTable) {
        KdPrint(("Can't get PCI IRQ Routing Table\n"));
        PciBios32_Deinit();
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    KdPrint(("PCI IRQ Routing Table size: %d(%#x) items\n", szIrqRoutingTable, szIrqRoutingTable));
    KdPrint(("PCI IRQ bitmap %#x\n", IrqBitMap));
    for(i=0; i<szIrqRoutingTable; i++) {
        KdPrint(("  Bus %d, Dev %x(%x) slot %d\n",
            IrqRoutingTable[i].BusId,
            IrqRoutingTable[i].DeviceNumber,
            IrqRoutingTable[i].DeviceNumber >> 3,
            IrqRoutingTable[i].SlotNumber
            ));
        KdPrint(("    INT#A Link %d, mask %x\n",
            IrqRoutingTable[i].INTA.Link,
            IrqRoutingTable[i].INTA.IRQBitMap
            ));
        KdPrint(("    INT#B Link %d, mask %x\n",
            IrqRoutingTable[i].INTB.Link,
            IrqRoutingTable[i].INTB.IRQBitMap
            ));
        KdPrint(("    INT#C Link %d, mask %x\n",
            IrqRoutingTable[i].INTC.Link,
            IrqRoutingTable[i].INTC.IRQBitMap
            ));
        KdPrint(("    INT#D Link %d, mask %x\n",
            IrqRoutingTable[i].INTD.Link,
            IrqRoutingTable[i].INTD.IRQBitMap
            ));
    }

    ExFreePool(IrqRoutingTable);
    IrqRoutingTable = NULL;
    //
    for(i=0; i<=IRP_MJ_MAXIMUM_FUNCTION; i++)
        driverObject->MajorFunction[i] = NULL;

    driverObject->DriverUnload = PciBios32_Deinit;
*/
    return STATUS_SUCCESS;

} // end PciBios32_Init()


VOID
PciBios32_Deinit(
    )
{
    if(mmMapped) {
        MmUnmapIoSpace(BIOS_VA, BASE_1MB_SIZE);
        BIOS_VA = NULL;
        mmMapped = FALSE;
    }
    BIOS_ver = -1;
    return;
} // end PciBios32_Deinit()

PBIOS32_SD
Check_BIOS32_SD(
    ULONG phAddr
    )
{
    PBIOS32_SD sd = (PBIOS32_SD)phAddr;
    ULONG i;
    UCHAR s;

    if(sd->sig[0] == '_' &&
       sd->sig[1] == '3' &&
       sd->sig[2] == '2' &&
       sd->sig[3] == '_' &&
       sd->length == 1 &&
       sd->rev    == 0) {
         // verify checksum
        for(s=0, i=0; i<16; i++) {
            s+=((PUCHAR)sd)[i];
        }
        if(s) {
            return NULL;
        }
        return sd;
    }
    return NULL;
} // end Check_BIOS32_SD()

PPCI_BIOS32_IRQ_ROUTING_TABLE_ENTRY
PciGetIrqRoutingTable(
    PULONG  pItemCount,
    PULONG  pIrqBitMap
    )
{
    PPCI_BIOS32_IRQ_ROUTING_TABLE_ENTRY t;
    IRQ_ROUTING_OPTIONS p;
    USHORT IrqBitMap = 0;

    t = (PPCI_BIOS32_IRQ_ROUTING_TABLE_ENTRY)ExAllocatePool(NonPagedPool, 64*1024);
    if(!t)
        return NULL;
    p.Addr = t;
    p.BufferSize = (USHORT)0xfffe;

    __asm {
        pushad
        xor   ebx,ebx
        xor   eax,eax
        mov   ax,cs
        push  eax
        lea   edi,[p]
        mov   eax, PCI_GET_IRQ_ROUTING_OPTIONS
        call  PCI_BIOS
        jc    no_pci_irt
        mov   IrqBitMap,bx
no_pci_irt:
        popad
    }
    if(IrqBitMap && p.BufferSize) {
        (*pItemCount) = p.BufferSize / sizeof(PCI_BIOS32_IRQ_ROUTING_TABLE_ENTRY);
        (*pIrqBitMap) = IrqBitMap;
        return t;
    }
    ExFreePool(t);
    return NULL;
} // end PciGetIrqRoutingTable()

ULONG
PciIsBIOSPresent(
    )
{
    ULONG ret_val = -1;
    __asm {
        pushad
        xor   ebx,ebx
        xor   eax,eax
        mov   ax,cs
        push  eax
        mov   eax, PCI_BIOS_PRESENT
        call  PCI_BIOS
        jc    no_pci_bios1
        cmp   edx,' ICP'  // reverse 'PCI '
        jne   no_pci_bios1
        and   ebx,0xffff
        mov   ret_val,ebx
no_pci_bios1:
        popad
    }
    return ret_val;
} // end PciIsBIOSPresent()

